@@ -27,6 +27,7 @@
 
 #include "common.h"
 #include "sys-tree.h"
+#include "freelist.h"
 #include "treetype.h"
 
 /* User to setuid to and directory to chroot to when we drop privs */
@@ -34,8 +35,9 @@
 # define PRIVDROP_USER		"nobody"
 #endif
 
-#define PRIVDROP_CHROOT_DIR	"/var/empty"
-
+#ifndef PRIVDROP_CHROOT_DIR
+# define PRIVDROP_CHROOT_DIR	"/var/empty"
+#endif
 /*
  * Capture length for libpcap: Must fit the link layer header, plus 
  * a maximally sized ip/ipv6 header and most of a TCP header
@@ -71,6 +73,13 @@ struct STATISTIC {
 #define TRACK_IP_PROTO		2	/* src/dst/proto 3-tuple */
 #define TRACK_IP_ONLY		3	/* src/dst tuple */
 
+/*
+ * This structure contains optional information carried by Option Data
+ * Record.
+ */
+struct OPTION {
+	uint32_t sample;
+};
 /*
  * This structure is the root of the flow tracking system.
  * It holds the root of the tree of active flows and the head of the
@@ -81,7 +90,11 @@ struct FLOWTRACK {
 	FLOW_HEAD(FLOWS, FLOW) flows;		/* Top of flow tree */
 	EXPIRY_HEAD(EXPIRIES, EXPIRY) expiries;	/* Top of expiries tree */
 
+	struct freelist flow_freelist;		/* Freelist for flows */
+	struct freelist expiry_freelist;	/* Freelist for expiry events */
+
 	unsigned int num_flows;			/* # of active flows */
+	unsigned int max_flows;			/* Max # of active flows */
 	u_int64_t next_flow_seq;		/* Next flow ID */
 
 	/* Stuff related to flow export */
@@ -100,6 +113,7 @@ struct FLOWTRACK {
 
 	/* Statistics */
 	u_int64_t total_packets;		/* # of good packets */
+	u_int64_t non_sampled_packets;		/* # of not sampled packets */
 	u_int64_t frag_packets;			/* # of fragmented packets */
 	u_int64_t non_ip_packets;		/* # of not-IP packets */
 	u_int64_t bad_packets;			/* # of bad packets */
@@ -129,6 +143,9 @@ struct FLOWTRACK {
 	u_int64_t expired_overbytes;
 	u_int64_t expired_maxflows;
 	u_int64_t expired_flush;
+
+	/* Optional information */
+	struct OPTION option;
 };
 
 /*
@@ -195,14 +212,14 @@ u_int32_t timeval_sub_ms(const struct timeval *t1, const struct timeval *t2);
 
 /* Prototypes for functions to send NetFlow packets, from netflow*.c */
 int send_netflow_v1(struct FLOW **flows, int num_flows, int nfsock,
-    u_int64_t *flows_exported, struct timeval *system_boot_time, 
-    int verbose_flag);
+    u_int16_t ifidx, u_int64_t *flows_exported, struct timeval *system_boot_time, 
+    int verbose_flag, struct OPTION *option);
 int send_netflow_v5(struct FLOW **flows, int num_flows, int nfsock,
-    u_int64_t *flows_exported, struct timeval *system_boot_time,
-    int verbose_flag);
+    u_int16_t ifidx, u_int64_t *flows_exported, struct timeval *system_boot_time,
+    int verbose_flag, struct OPTION *option);
 int send_netflow_v9(struct FLOW **flows, int num_flows, int nfsock,
-    u_int64_t *flows_exported, struct timeval *system_boot_time,
-    int verbose_flag);
+    u_int16_t ifidx, u_int64_t *flows_exported, struct timeval *system_boot_time,
+    int verbose_flag, struct OPTION *option);
 
 /* Force a resend of the flow template */
 void netflow9_resend_template(void);
